      program aq_driver_hv
c****************************************************************************
c     purpose: Setting up calling sequence and Allocate memories.
c******************************************************************************
c---+----1----+----2----+----3----+----4----+----5----+----6----+----7-----
      use HVParallelCommunication
c      implicit none
c--------------------------------------------------------------------------
      include 'aqmax.param'
      include 'aqms.param'
      include 'aqcon1.cmm'
      include 'aqcon2.cmm'
      include 'aqcon5.cmm'
      include 'aqcont.cmm'
      include 'aqfile.cmm'  
      include 'aqrxn.cmm'	
      include 'aqsymb.cmm'
      include 'aqspec.cmm'
      include 'aqindx.cmm'

c--------------------------------------------------------------------------
      integer :: ix=ixm, iy=iym, iz=izm  ! Grid Dimensions
      integer :: N_gas=iLm               ! No. of gas species
      integer :: N_liquid=iLm            ! No. of liquid species
      integer :: N_particle=iLptm        ! No. of particulate species
c--------------------------------------------------------------------------
c      dimension sg1(ixm*iym*izm*iLm),sl1(ixm*iym*izm*iLm)
c      dimension sp1(ixm*iym*izm*ilptm)
      real, dimension(:,:,:,:), pointer :: sg1, sg1_h, sg1_v
      real, dimension(:,:,:,:), pointer :: sl1, sl1_h, sl1_v 
      real, dimension(:,:,:,:), pointer :: sp1, sp1_h, sp1_v 
c      dimension u(ixm*iym*izm),v(ixm*iym*izm),w(ixm*iym*izm)
      real, dimension(:,:,:), pointer :: u, u_h, u_v
      real, dimension(:,:,:), pointer :: v, v_h, v_v 
      real, dimension(:,:,:), pointer :: w, w_h, w_v
c      dimension kh(ixm*iym*izm),kv(ixm*iym*izm),t(ixm*iym*izm)
      real, dimension(:,:,:), pointer :: kh, kh_h, kh_v 
      real, dimension(:,:,:), pointer :: kv, kv_h, kv_v
      real, dimension(:,:,:), pointer :: t,  t_h,  t_v
c      dimension dz(ixm*iym*izm)
      real, dimension(:,:,:), pointer :: dz, dz_h, dz_v
c      dimension wc(ixm*iym*izm),wr2(ixm*iym*izm)
      real, dimension(:,:,:), pointer :: wc, wc_h, wc_v
      real, dimension(:,:,:), pointer :: wr2, wr_h, wr_v
c      dimension sprc(ixm*iym),rvel(ixm*iym*izm)
      real, dimension(:,:), pointer   :: sprc, sprc_h, sprc_v
      real, dimension(:,:,:), pointer :: rvel, rvel_h, rvel_v
c      dimension sx(iym*izm*2*iLm),sy(ixm*izm*2*iLm),sz(ixm*iym*iLm)            
      real, dimension(:,:,:,:), pointer :: sx, sx_h, sx_v
      real, dimension(:,:,:,:), pointer :: sy, sy_h, sy_v
      real, dimension(:,:,:),   pointer :: sz, sz_h, sz_v
c      dimension q(ixm*iym*iLm)
      real, dimension(:,:,:),   pointer :: q, q_h, q_v
c      dimension em(ixm*iym*izm*iLm)
      real, dimension(:,:,:,:), pointer :: em, em_h, em_v
c      dimension vg(ixm*iym*iLm),fz(ixm*iym*iLm)
      real, dimension(:,:,:),   pointer :: vg, vg_h, vg_v
      real, dimension(:,:,:),   pointer :: fz, fz_h, fz_v
c      dimension hdz(ixm*iym*izm)
      real, dimension(:,:,:),   pointer :: hdz, hdz_h, hdz_v
c      dimension h(ixm*iym),tlon(ixm*iym),tlat(ixm*iym)
      real, dimension(:,:),     pointer :: h,    h_h,    h_v
      real, dimension(:,:),     pointer :: tlon, tlon_h, tlon_v
      real, dimension(:,:),     pointer :: tlat, tlat_h, tlat_v
C      
      real   :: dx(ixm),dy(iym)
      real   :: sigmaz(izm)
      integer :: iunit(25),iout(25),idate(3)
      real   :: dht, baseh, ut, dt, dtmax
      integer :: irxng, irxnl, ixtrn, iytrn, iztrn
      integer :: iter, it, num, iend, mdt
c-------------------------------------------------------------------------------
c   Local Sizes
c-------------------------------------------------------------------------------
      integer :: izloc, icloc, ibounds(6)

c-------------------------------------------------------------------------------
c   MPI-realated declarations
c-------------------------------------------------------------------------------
      integer :: status(MPI_STATUS_SIZE)
      integer :: ierr, rc
      double precision :: StartTime, EndTime
c                                                                     
c-------------------------------------------------------------------------------
c   Initialize MPI Stuff
c-------------------------------------------------------------------------------
      call MPI_INIT( ierr )
      call MPI_COMM_RANK( MPI_COMM_WORLD, MyId, ierr )
      call MPI_COMM_SIZE( MPI_COMM_WORLD, Nprocs, ierr )
c-------------------------------------------------------------------------------
c    The data mapping scheme, special comm patterns, initialization
c-------------------------------------------------------------------------------
      call init_hv(ix,iy,iz,izloc,icloc,N_gas)

c-------------------------------------------------------------------------------
c    Check if not enough or too many workers
c-------------------------------------------------------------------------------
      if (Nprocs<=1) then
         print*,'There are only ',Nprocs-1,' worker(s),'
         print*,'   not enough for parallelization!'
	 goto 10
      else if (Nprocs-1>max(iz,ix*iy)) then
         print*,'There are ',Nprocs-1,' workers ',
     &        'and only ',iz,' h-slices, ',ix*iy,' v-columns.'
         print*,'Some workers will go to unemployment !' 
      end if	 
                                                                      
c----------------------------------------------------------------------c
c                      Allocate memory                                 c
c----------------------------------------------------------------------c
      if (Master) then
       call MemAlloc( ix, iy, iz, N_gas, N_liquid, N_particle, 
     &                 sg1, sl1, sp1, u, v, w, kh, kv, 
     &                 t, dz, wc, wr2, sprc, rvel, 
     &                 sx, sy, sz, q, em, vg, fz, 
     &                 hdz, h, tlon, tlat )
       call MinimAlloc( 
     &                 sg1_v, sl1_v, sp1_v, u_v, v_v, w_v, kh_v, kv_v, 
     &                 t_v, dz_v, wc_v, wr_v, sprc_v, rvel_v, 
     &                 sx_v, sy_v, sz_v, q_v, em_v, vg_v, fz_v, 
     &                 hdz_v, h_v, tlon_v, tlat_v )
       call MinimAlloc(
     &                 sg1_h, sl1_h, sp1_h, u_h, v_h, w_h, kh_h, kv_h, 
     &                 t_h, dz_h, wc_h, wr_h, sprc_h, rvel_h, 
     &                 sx_h, sy_h, sz_h, q_h, em_h, vg_h, fz_h, 
     &                 hdz_h, h_h, tlon_h, tlat_h )
      else if (HWorker .or. VWorker) then
       call MinimAlloc( sg1, sl1, sp1, u, v, w, kh, kv, 
     &                 t, dz, wc, wr2, sprc, rvel, 
     &                 sx, sy, sz, q, em, vg, fz, 
     &                 hdz, h, tlon, tlat )
       call MemAlloc( 1, icloc, iz, N_gas, N_liquid, N_particle, 
     &                 sg1_v, sl1_v, sp1_v, u_v, v_v, w_v, kh_v, kv_v, 
     &                 t_v, dz_v, wc_v, wr_v, sprc_v, rvel_v, 
     &                 sx_v, sy_v, sz_v, q_v, em_v, vg_v, fz_v, 
     &                 hdz_v, h_v, tlon_v, tlat_v )
       call MemAlloc( ix, iy, izloc, N_gas, N_liquid, N_particle, 
     &                 sg1_h, sl1_h, sp1_h, u_h, v_h, w_h, kh_h, kv_h, 
     &                 t_h, dz_h, wc_h, wr_h, sprc_h, rvel_h, 
     &                 sx_h, sy_h, sz_h, q_h, em_h, vg_h, fz_h, 
     &                 hdz_h, h_h, tlon_h, tlat_h )
      else
       print*,'I am neither Master nor Worker ...'
      endif ! Master
c----------------------------------------------------------------------c
                                                                      
c-------------------------------------------------------------------------------
c                       Start the timer
c-------------------------------------------------------------------------------
      StartTime = MPI_WTIME()
                                                                            
c----------------------------------------------------------------------c
c                  Master Performs Setup Simulation                                 c           
c----------------------------------------------------------------------c
      if (Master) then
      T1 = MPI_WTIME()
        call aq_setp(ix,iy,iz,numl,nbin,ixtrn,iytrn,iztrn,
     1                irxng,irxnl,ixm,iym,izm,ilm,
     2                dx,dy,dht,
     2                sigmaz,baseh,
     3                sg1,sl1,sp1,
     3                u,v,w,kh,kv,t,wc,wr2,sprc,q,em,
     4                vg,fz,sx,sy,sz)
        endif
        !call  int_distrib1(iend) 
        !if (irxng.eq.0) then
	 !  call aq_set_jac(numl(1,2))
	!endif   
        if(Master)then
        num=numl(1,1)
        call input1(idate,ut,iend,ix,iy,iz,
     1	             sg1,sl1,sp1,
     2               tlon,tlat,h,hdz,dx,dy,dz,dht,
     3               sigmaz,baseh,dtmax,iunit,iout)
        call prtini
      endif ! Master
 
      call int_distrib1(iend,numl)
      if (irxng.eq.0) then
         call aq_set_jac(numl(1,2))
      endif   
      T2 = MPI_WTIME()
c----------------------------------------------------------------------c
c                   Simulation begins here                             c
c----------------------------------------------------------------------c
 
      do 10 it=int(ut)+1,iend
c      
      if (Master) then
      T3 = MPI_WTIME()
c      
        call input2(ix,iy,iz,num,it,idate,
     &               sg1,u,v,w,kh,kv,t,
     &		     wc,wr2,rvel,q,em,vg,fz,sprc,
     &               sx,sy,sz,dx,dy,hdz,iunit)
c
        if (it.eq.1) then
          call input_check(ix,iy,iz,numl(1,2),sg1,
     &                      ixtrn,iytrn,iztrn,irxng)
        endif  ! (it.eq.1) 
     
        call aq_hori_dt_Strang(ix,iy,iz,ixtrn,iytrn,
     &                 u,v,dx,dy,dtmax,dt,mdt)
c     
      T4 = MPI_WTIME()
      endif  !  Master
      
      
c-----------------------------------------------------------------------
c     Master Distributes data 
c-----------------------------------------------------------------------
c      
      T5 = MPI_WTIME()
      call int_distrib(numl,nbin,ixtrn,iytrn,iztrn,
     &                  irxng,irxnl,num,mdt,idate,iend)
      call real_distrib(ix,iy,iz,is,dx,dy,sigmaz,
     &                  dht,baseh,rmw,dt,ut) 
      call cmm_distrib() 
     
     
      call distrib_hv(ix,iy,iz,izloc,icloc,
     &                    N_gas, N_liquid, N_particle,
     &                    sg1, sg1_h, sg1_v,
     &                    sl1, sl1_h, sl1_v,
     &                    sp1, sp1_h, sp1_v,
     &                    u, v, w, u_h, v_h, w_v,
     &                    kh, kv, t, kh_h, kv_v, t_v,
     &                    wc, wc_v, wr2, wr_v,
     &                    rvel, rvel_v,
     &                    q, q_v, sprc, sprc_v,
     &                    em, em_v,
     &                    vg, vg_v, fz, fz_v,
     &                    hdz, hdz_v,
     &                    sx, sx_h, sy, sy_h, sz, sz_v,
     &                    h, h_v, tlon, tlon_v,
     &                    tlat, tlat_v, dz, dz_v )    
c 
      T6 = MPI_WTIME()
 
      
c-----------------------------------------------------------------------
c     Start Small time-split steps 
c-----------------------------------------------------------------------

      do 30 iter=1,mdt
c
c      
      if (HWorker) then
	 ibounds=(/1,ix,1,iy,1,no_of_hslices(MyId)/)
         if (ixtrn*iytrn*iztrn.eq.0) then
            call conv_conc(ix*iy*izloc,num,sg1_h,
     &                   sg1_h(1,1,1,iair),rmw,iair,8)
         endif ! ixtrn*iytrn*iztrn.eq.0
         if (ixtrn.eq.0) then
	    call tranx(ix,iy,izloc,ibounds,
     &	                num,sg1_h,u_h,kh_h,sx_h,dt,dx)
         endif ! ixtrn.eq.0
         if (iytrn.eq.0) then ! Y-Transport
	    call trany(ix,iy,izloc,ibounds,
     &	                num,sg1_h,v_h,kh_h,sy_h,dt,dy)
         endif ! iytrn.eq.0
       endif ! HWorker
       
 
c      T3 = MPI_WTIME()
c  ----------------------------------------------------------------------------------------------
c              Change data format from x-slices to y-slices 
         call shuffle_h2v_4D(ix,iy,iz,izloc,icloc,N_gas,sg1_h,sg1_v)  
c------------------------------------------------------------------------------------------------- 
c      T4 = MPI_WTIME()
	 	 
      if (VWorker) then
	 ibounds=(/1,1,1,no_of_vcols(myid),1,iz/)
         if (iztrn.eq.0) then 
            call tranz(1,icloc,iz,ibounds,
     &	                num,sg1_v,w_v,kv_v,q_v,
     &                  em_v,vg_v,sz_v,dt,dz_v)
         endif ! iztrn.eq.0
         if (ixtrn*iytrn*iztrn.eq.0) then
            call conv_conc(icloc*iz,num,sg1_v,
     &                   sg1_v(1,1,1,iair),rmw,iair,7)
         endif ! ixtrn*iytrn*iztrn.eq.0
c      
         if (irxng.eq.0) then 
           call rxn(1,icloc,iz,
     &               ibounds,
     &               idate,ut,2*dt,tlon_v,tlat_v,
     &               h_v,hdz_v,sg1_v,sl1_v,sp1_v,t_v,
     &               wc_v,wr_v,sprc_v,rvel_v)
         endif ! (irxng.eq.0)
c
         if (ixtrn*iytrn*iztrn.eq.0) then
           call conv_conc(icloc*iz,num,sg1_v,
     &                     sg1_v(1,1,1,iair),rmw,iair,8)
         endif ! (ixtrn*iytrn*iztrn.eq.0)
c	 
         if (iztrn.eq.0) then 
            call tranz(1,icloc,iz,
     &                  ibounds,
     &                  num,sg1_v,w_v,kv_v,q_v,
     &                  em_v,vg_v,sz_v,dt,dz_v)
         endif ! (iztrn.eq.0)
c	 
c	 
       endif ! VWorker
       	 
c      T5 = MPI_WTIME()
c  ----------------------------------------------------------------------------------------------
c              Change data format from y-slices to x-slices 
         call shuffle_v2h_4D(ix,iy,iz,izloc,icloc,N_gas,sg1_h,sg1_v)  
c  ----------------------------------------------------------------------------------------------
c      T6 = MPI_WTIME()

      if (HWorker) then
	ibounds=(/1,ix,1,iy,1,no_of_hslices(MyId)/)
         if (iytrn.eq.0) then 
	    call trany(ix,iy,izloc,ibounds,
     &	                num,sg1_h,v_h,kh_h,sy_h,dt,dy)
         endif ! (iytrn.eq.0)

         if (ixtrn.eq.0) then
	     call tranx(ix,iy,izloc,ibounds,
     &	                 num,sg1_h,u_h,kh_h,sx_h,dt,dx)
         endif ! (ixtrn.eq.0)
         if (ixtrn*iytrn*iztrn.eq.0) then
             call conv_conc(ix*iy*izloc,num,sg1_h,
     &                   sg1_h(1,1,1,iair),rmw,iair,7)
         endif ! ixtrn*iytrn*iztrn.eq.0
      endif ! HWorker

      ut=ut+2.d0*dt/3600.0
      if (Master) then
        call aq_clock_n(idate,ut)
        write(6,*) 'idate=',idate,'   ut=',ut
      endif ! Master	
c
30    continue
c------------------------------------------------------------------------
c    End small time-split steps
c------------------------------------------------------------------------

c------------------------------------------------------------------------
c    Master Process Gets Concentration Data From Workers
c------------------------------------------------------------------------
c      T7 = MPI_WTIME()
      call gather_h_4D(ix,iy,iz,izloc,N_gas,sg1,sg1_h)
c      T8 = MPI_WTIME()
c      call gather_h_4D(ix,iy,iz,izloc,N_liquid,sl1,sl1_h)
c      call gather_h_4D(ix,iy,iz,izloc,N_particle,sp1,sp1_h)
c------------------------------------------------------------------------
c    Master Process Prints Data
c------------------------------------------------------------------------
      if (Master) then
      T7 = MPI_WTIME()
        if (mod(it,iprnt).eq.0) then
           call prtout(idate,ut,ix,iy,iz,sg1,sl1,iout)
        endif
        if (mod(it,iscrat).eq.0) then 
           call prtemp(idate,ut,ix,iy,iz,sg1,sl1,iout)
        endif
      T8 = MPI_WTIME()
      endif
c     
10    continue  ! end of Big Loop
c-----------------------------------------------------------------------c
c                      Simulation ends here                             c
c-----------------------------------------------------------------------c

c      call close_ioap
 100  continue
c-------------------------------------------------------------------------------
c    Stop the timer
c-------------------------------------------------------------------------------
      EndTime = MPI_WTIME()-StartTime
      if (Master) then
         print*,'Master prints ascii files ...'
c         call plot3d(Nprocs,EndTime,ut,ix,iy,iz,iLm,ixtrn,
c     &                iytrn,iztrn,irxng,sg1)
         open(unit=98,file='/scratch/pmiehe/output/hvpartimes.m',
     &              action='WRITE',position='APPEND');
         write(98,*) 'Nprocs=',Nprocs,'  CPU_TIME=',EndTime,' sec.'
         write(98,*) T2-T1,T4-T3,T6-T5,T8-T7
	 close(98)
      endif
      print("('Proc[',I2,'] Time = ',F9.2,' sec. = ',F7.2,' min.')"),
     &      MyId,EndTime,EndTime/60.0    
c        
      call MPI_FINALIZE( Ierr )
c      
      end program aq_driver_hv



c********************************************************************
      subroutine cho_tmp(ix,iy,iz,num,sg1,sx)
c********************************************************************
      parameter (mxspg=250)
      include 'aqindx.cmm'
      dimension sg1(ix,iy,iz,*),sx(iy,iz,2,*)
      do i=1,ix
      do j=1,iy
      do k=1,iz
      sg1(i,j,k,7)=40.*sg1(i,j,k,iair)*1.e-09
      enddo
      enddo
      enddo
      return
      end


      subroutine plot3d(Np,cpu,ut,ix,iy,iz,is,
     &                ixtrn,iytrn,iztrn,irxng,sg1)
      integer Np,ix,iy,iz,is,ixtrn,iytrn,iztrn,irxng
      real sg1(ix,iy,iz,is)
      double precision cpu
            integer ilayer
      character(len=4) :: c
      do ilayer=1,iz!,iz-1
      write(c,10) 1000+ilayer
      open(10,file=
     & '/scratch/pmiehe/output/parpgf_lyr' // c(3:4) // '.m')
      write(10,*) '%******** Parallel Symmetric Code **************'
      write(10,*) 'Nprocs=',Np,';'     
      write(10,*) 'P_CPU_TIME=',cpu,';'    
      write(10,*) 'P_ut=',ut,';'     
      write(10,*) 'P_ixtrn=',ixtrn,';'
      write(10,*) 'P_iytrn=',iytrn,';'
      write(10,*) 'P_iztrn=',iztrn,';'
      write(10,*) 'P_irxng=',irxng,';'
      write(10,*) 'P_PGF=['
      write(10,11) (((sg1(i,j,ilayer,l),i=1,ix),j=1,iy),l=1,27)
      write(10,*) '];'
      write(10,*) '%***********************************'
      close(10)
      enddo
 10   format(I4)
 11   format(9(E16.8,1X))     
      return
      end
