      program aq_driver_xy
c****************************************************************************
c     purpose: Setting up calling sequence and Allocate memories.
c******************************************************************************
c---+----1----+----2----+----3----+----4----+----5----+----6----+----7-----
      use XYParallelCommunication
c      implicit none
c--------------------------------------------------------------------------
      include 'aqmax.param'
      include 'aqms.param'
      include 'aqcon1.cmm'
      include 'aqcon2.cmm'
      include 'aqcon5.cmm'
      include 'aqcont.cmm'
      include 'aqfile.cmm'  
      include 'aqrxn.cmm'	
      include 'aqsymb.cmm'
      include 'aqspec.cmm'
      include 'aqindx.cmm'

c--------------------------------------------------------------------------
      integer :: ix=ixm, iy=iym, iz=izm  ! Grid Dimensions
      integer :: N_gas=iLm               ! No. of gas species
      integer :: N_liquid=iLm            ! No. of liquid species
      integer :: N_particle=iLptm        ! No. of particulate species
c--------------------------------------------------------------------------
c      dimension sg1(ixm*iym*izm*iLm),sl1(ixm*iym*izm*iLm)
c      dimension sp1(ixm*iym*izm*ilptm)
      real, dimension(:,:,:,:), pointer :: sg1, sg1_x, sg1_y
      real, dimension(:,:,:,:), pointer :: sl1, sl1_x, sl1_y 
      real, dimension(:,:,:,:), pointer :: sp1, sp1_x, sp1_y 
c      dimension u(ixm*iym*izm),v(ixm*iym*izm),w(ixm*iym*izm)
      real, dimension(:,:,:), pointer :: u, u_x, u_y
      real, dimension(:,:,:), pointer :: v, v_x, v_y 
      real, dimension(:,:,:), pointer :: w, w_x, w_y
c      dimension kh(ixm*iym*izm),kv(ixm*iym*izm),t(ixm*iym*izm)
      real, dimension(:,:,:), pointer :: kh, kh_x, kh_y 
      real, dimension(:,:,:), pointer :: kv, kv_x, kv_y
      real, dimension(:,:,:), pointer :: t,  t_x,  t_y
c      dimension dz(ixm*iym*izm)
      real, dimension(:,:,:), pointer :: dz, dz_x, dz_y
c      dimension wc(ixm*iym*izm),wr2(ixm*iym*izm)
      real, dimension(:,:,:), pointer :: wc, wc_x, wc_y
      real, dimension(:,:,:), pointer :: wr2, wr_x, wr_y
c      dimension sprc(ixm*iym),rvel(ixm*iym*izm)
      real, dimension(:,:), pointer   :: sprc, sprc_x, sprc_y
      real, dimension(:,:,:), pointer :: rvel, rvel_x, rvel_y
c      dimension sx(iym*izm*2*iLm),sy(ixm*izm*2*iLm),sz(ixm*iym*iLm)            
      real, dimension(:,:,:,:), pointer :: sx, sx_x, sx_y 
      real, dimension(:,:,:,:), pointer :: sy, sy_x, sy_y
      real, dimension(:,:,:),   pointer :: sz, sz_x, sz_y
c      dimension q(ixm*iym*iLm)
      real, dimension(:,:,:),   pointer :: q, q_x, q_y
c      dimension em(ixm*iym*izm*iLm)
      real, dimension(:,:,:,:), pointer :: em, em_x, em_y
c      dimension vg(ixm*iym*iLm),fz(ixm*iym*iLm)
      real, dimension(:,:,:),   pointer :: vg, vg_x, vg_y
      real, dimension(:,:,:),   pointer :: fz, fz_x, fz_y
c      dimension hdz(ixm*iym*izm)
      real, dimension(:,:,:),   pointer :: hdz, hdz_x, hdz_y
c      dimension h(ixm*iym),tlon(ixm*iym),tlat(ixm*iym)
      real, dimension(:,:),     pointer :: h,    h_x,    h_y
      real, dimension(:,:),     pointer :: tlon, tlon_x, tlon_y 
      real, dimension(:,:),     pointer :: tlat, tlat_x, tlat_y
C      
      real   :: dx(ixm),dy(iym)
      real   :: sigmaz(izm)
      integer :: iunit(25),iout(25),idate(3)
      real   :: dht, baseh, ut, dt, dtmax
      integer :: irxng, irxnl, ixtrn, iytrn, iztrn
      integer :: iter, it, num, iend, mdt
c-------------------------------------------------------------------------------
c   Local Sizes
c-------------------------------------------------------------------------------
      integer :: ixloc, iyloc, ibounds(6)

c-------------------------------------------------------------------------------
c   MPI-realated declarations
c-------------------------------------------------------------------------------
      integer :: status(MPI_STATUS_SIZE)
      integer :: ierr, rc
      double precision :: StartTime, EndTime
c                                                                     
c-------------------------------------------------------------------------------
c   Initialize MPI Stuff
c-------------------------------------------------------------------------------
      call MPI_INIT( ierr )
      call MPI_COMM_RANK( MPI_COMM_WORLD, MyId, ierr )
      call MPI_COMM_SIZE( MPI_COMM_WORLD, Nprocs, ierr )
c-------------------------------------------------------------------------------
c    The data mapping scheme, communciation pattern, initialization
c-------------------------------------------------------------------------------
      call init_xy(ix,iy,iz,ixloc,iyloc,N_gas)
c-------------------------------------------------------------------------------
c    Check if not enough or too many workers
c-------------------------------------------------------------------------------
      if (Nprocs<=1) then
         print*,'There are only ',Nprocs-1,' worker(s),'
         print*,'   not enough for parallelization!'
	 goto 10
      else if (Nprocs-1>max(ix,iy)) then
         print*,'There are ',Nprocs-1,' workers ',
     &        'and only ',ix,' y-slices, ',iy,' x-slices.'
         print*,'Some workers will go to unemployment !' 
      end if	 
                                                                      
c----------------------------------------------------------------------c
c                      Allocate memory                                 c
c----------------------------------------------------------------------c
      if (Master) then
       call MemAlloc( ix, iy, iz, N_gas, N_liquid, N_particle, 
     &                 sg1, sl1, sp1, u, v, w, kh, kv, 
     &                 t, dz, wc, wr2, sprc, rvel, 
     &                 sx, sy, sz, q, em, vg, fz, 
     &                 hdz, h, tlon, tlat )
       call MinimAlloc( 
     &                 sg1_y, sl1_y, sp1_y, u_y, v_y, w_y, kh_y, kv_y, 
     &                 t_y, dz_y, wc_y, wr_y, sprc_y, rvel_y, 
     &                 sx_y, sy_y, sz_y, q_y, em_y, vg_y, fz_y, 
     &                 hdz_y, h_y, tlon_y, tlat_y )
       call MinimAlloc(
     &                 sg1_x, sl1_x, sp1_x, u_x, v_x, w_x, kh_x, kv_x, 
     &                 t_x, dz_x, wc_x, wr_x, sprc_x, rvel_x, 
     &                 sx_x, sy_x, sz_x, q_x, em_x, vg_x, fz_x, 
     &                 hdz_x, h_x, tlon_x, tlat_x )
      else if (XWorker .or. YWorker) then
       call MinimAlloc( sg1, sl1, sp1, u, v, w, kh, kv, 
     &                 t, dz, wc, wr2, sprc, rvel, 
     &                 sx, sy, sz, q, em, vg, fz, 
     &                 hdz, h, tlon, tlat )
       call MemAlloc( ixloc, iy, iz, N_gas, N_liquid, N_particle, 
     &                 sg1_y, sl1_y, sp1_y, u_y, v_y, w_y, kh_y, kv_y, 
     &                 t_y, dz_y, wc_y, wr_y, sprc_y, rvel_y, 
     &                 sx_y, sy_y, sz_y, q_y, em_y, vg_y, fz_y, 
     &                 hdz_y, h_y, tlon_y, tlat_y )
       call MemAlloc( ix, iyloc, iz, N_gas, N_liquid, N_particle, 
     &                 sg1_x, sl1_x, sp1_x, u_x, v_x, w_x, kh_x, kv_x, 
     &                 t_x, dz_x, wc_x, wr_x, sprc_x, rvel_x, 
     &                 sx_x, sy_x, sz_x, q_x, em_x, vg_x, fz_x, 
     &                 hdz_x, h_x, tlon_x, tlat_x )
      else
       print*,'I am neither Master nor Worker ...'
      endif ! Master
c----------------------------------------------------------------------c
                                                                      
c-------------------------------------------------------------------------------
c                       Start the timer
c-------------------------------------------------------------------------------
      StartTime = MPI_WTIME()
                                                                            
c----------------------------------------------------------------------c
c                  Master Performs Setup Simulation                                 c           
c----------------------------------------------------------------------c
      if (Master) then
      T1 = MPI_WTIME()
        call aq_setp(ix,iy,iz,numl,nbin,ixtrn,iytrn,iztrn,
     1                irxng,irxnl,ixm,iym,izm,ilm,
     2                dx,dy,dht,
     2                sigmaz,baseh,
     3                sg1,sl1,sp1,
     3                u,v,w,kh,kv,t,wc,wr2,sprc,q,em,
     4                vg,fz,sx,sy,sz)
        endif
        !call  int_distrib1(iend) 
        !if (irxng.eq.0) then
	 !  call aq_set_jac(numl(1,2))
	!endif   
        if(Master)then
        num=numl(1,1)
        call input1(idate,ut,iend,ix,iy,iz,
     1	             sg1,sl1,sp1,
     2               tlon,tlat,h,hdz,dx,dy,dz,dht,
     3               sigmaz,baseh,dtmax,iunit,iout)
        call prtini
      endif ! Master
 
      call int_distrib1(iend,numl)
      if (irxng.eq.0) then
         call aq_set_jac(numl(1,2))
      endif   
      T2 = MPI_WTIME()
c----------------------------------------------------------------------c
c                   Simulation begins here                             c
c----------------------------------------------------------------------c
 
      do 10 it=int(ut)+1,iend
c      
c
      if (Master) then
      T3 = MPI_WTIME()
c      
        call input2(ix,iy,iz,num,it,idate,
     &               sg1,u,v,w,kh,kv,t,
     &		     wc,wr2,rvel,q,em,vg,fz,sprc,
     &               sx,sy,sz,dx,dy,hdz,iunit)
c
        if (it.eq.1) then
          call input_check(ix,iy,iz,numl(1,2),sg1,
     &                      ixtrn,iytrn,iztrn,irxng)
        endif  ! (it.eq.1) 
     
        call aq_hori_dt_Strang(ix,iy,iz,ixtrn,iytrn,
     &                 u,v,dx,dy,dtmax,dt,mdt)
c     
      T4 = MPI_WTIME()
      endif  !  Master
      
      
c-----------------------------------------------------------------------
c     Master Distributes data 
c-----------------------------------------------------------------------
c      
      T5 = MPI_WTIME()
      call int_distrib(numl,nbin,ixtrn,iytrn,iztrn,
     &                  irxng,irxnl,num,mdt,idate,iend)
      call real_distrib(ix,iy,iz,is,dx,dy,sigmaz,
     &                  dht,baseh,rmw,dt,ut) 
      call cmm_distrib() 
     
     
      call distrib_xy(ix,iy,iz,ixloc,iyloc,
     &                    N_gas, N_liquid, N_particle,
     &                    sg1, sg1_x, sg1_y,
     &                    sl1, sl1_x, sl1_y,
     &                    sp1, sp1_x, sp1_y,
     &                    u, v, w, u_x, v_x, w_x, u_y, v_y, w_y,
     &                    kh, kv, t, kh_x, kv_x, t_x, kh_y, kv_y, t_y,
     &                    wc, wc_x, wc_y, wr2, wr_x, wr_y, 
     &                    rvel, rvel_x, rvel_y,
     &                    q, q_x, q_y, sprc, sprc_x, sprc_y,    
     &                    em, em_x, em_y,     
     &                    vg, vg_x, vg_y, fz, fz_x, fz_y,     
     &                    hdz, hdz_x, hdz_y, 
     &                    sx, sx_x, sy, sy_y, sz, sz_x, sz_y,
     &                    h,   h_x,  h_y, tlon, tlon_x, tlon_y,
     &                    tlat, tlat_x, tlat_y, dz, dz_x, dz_y  )    
c 
      T6 = MPI_WTIME()
 
      
c-----------------------------------------------------------------------
c     Start Small time-split steps 
c-----------------------------------------------------------------------

      do 30 iter=1,mdt
c
      if (XWorker) then
	ibounds=(/1,ix,1,no_of_xslices(MyId),1,iz/)
         if (ixtrn*iytrn*iztrn.eq.0) then
            call conv_conc(ix*iyloc*iz,num,sg1_x,
     &                   sg1_x(1,1,1,iair),rmw,iair,8)
         endif ! ixtrn*iytrn*iztrn.eq.0
         if (ixtrn.eq.0) then
	    call tranx(ix,iyloc,iz,ibounds,
     &	                num,sg1_x,u_x,kh_x,sx_x,dt,dx)
         endif ! ixtrn.eq.0
       endif ! XWorker
       
 
c      T3 = MPI_WTIME()
c  ----------------------------------------------------------------------------------------------
c              Change data format from x-slices to y-slices 
	 if (Master) print*,'x2y'
         call shuffle_x2y_4D(ix,iy,iz,ixloc,iyloc,N_gas,sg1_x,sg1_y)  
c------------------------------------------------------------------------------------------------- 
c      T4 = MPI_WTIME()

	 	 
      if (YWorker) then
	ibounds = (/ 1, no_of_yslices(MyId), 1, iy, 1, iz /)
         if (iytrn.eq.0) then ! Y-Transport
	    call trany(ixloc,iy,iz,ibounds,
     &	                num,sg1_y,v_y,kh_y,sy_y,dt,dy)
         endif ! iytrn.eq.0
         if (iztrn.eq.0) then 
            call tranz(ixloc,iy,iz,ibounds,
     &	                num,sg1_y,w_y,kv_y,q_y,
     &                  em_y,vg_y,sz_y,dt,dz_y)
         endif ! iztrn.eq.0
	 
         if (ixtrn*iytrn*iztrn.eq.0) then
            call conv_conc(ixloc*iy*iz,num,sg1_y,
     &                   sg1_y(1,1,1,iair),rmw,iair,7)
         endif ! ixtrn*iytrn*iztrn.eq.0
c      
         if (irxng.eq.0) then 
           call rxn(ixloc,iy,iz,ibounds,
     &               idate,ut,2*dt,tlon_y,tlat_y,
     &               h_y,hdz_y,sg1_y,sl1_y,sp1_y,t_y,
     &               wc_y,wr_y,sprc_y,rvel_y)
         endif ! (irxng.eq.0)
c
         if (ixtrn*iytrn*iztrn.eq.0) then
           call conv_conc(ixloc*iy*iz,num,sg1_y,
     &                     sg1_y(1,1,1,iair),rmw,iair,8)
         endif ! (ixtrn*iytrn*iztrn.eq.0)
c	 
         if (iztrn.eq.0) then 
            call tranz(ixloc,iy,iz,ibounds,
     &                  num,sg1_y,w_y,kv_y,q_y,
     &                  em_y,vg_y,sz_y,dt,dz_y)
         endif ! (iztrn.eq.0)
c	 
         if (iytrn.eq.0) then 
	    call trany(ixloc,iy,iz,ibounds,
     &	                num,sg1_y,v_y,kh_y,sy_y,dt,dy)
         endif ! (iytrn.eq.0)
c	 
       endif ! YWorker
       	 
c      T5 = MPI_WTIME()
c  ----------------------------------------------------------------------------------------------
c              Change data format from y-slices to x-slices 
	 if (Master) print*,'y2x'
         call shuffle_y2x_4D(ix,iy,iz,ixloc,iyloc,N_gas,sg1_x,sg1_y)  
c  ----------------------------------------------------------------------------------------------
c      T6 = MPI_WTIME()

      if (XWorker) then
	ibounds=(/1,ix,1,no_of_xslices(MyId),1,iz/)
         if (ixtrn.eq.0) then
	     call tranx(ix,iyloc,iz,ibounds,
     &	                 num,sg1_x,u_x,kh_x,sx_x,dt,dx)
         endif ! (ixtrn.eq.0)
         if (ixtrn*iytrn*iztrn.eq.0) then
             call conv_conc(ix*iyloc*iz,num,sg1_x,
     &                   sg1_x(1,1,1,iair),rmw,iair,7)
         endif ! ixtrn*iytrn*iztrn.eq.0
      endif ! XWorker

      ut=ut+2.d0*dt/3600.0
      if (Master) then
        call aq_clock_n(idate,ut)
        write(6,*) 'idate=',idate,'   ut=',ut
      endif ! Master	
c
30    continue
c------------------------------------------------------------------------
c    End small time-split steps
c------------------------------------------------------------------------

c------------------------------------------------------------------------
c    Master Process Gets Concentration Data From Workers
c------------------------------------------------------------------------
c      T7 = MPI_WTIME()
      call gather_x_4D(ix,iy,iz,iyloc,N_gas,sg1,sg1_x)
c      T8 = MPI_WTIME()
c      call gather_x_4D(ix,iy,iz,iyloc,N_liquid,sl1,sl1_x)
c      call gather_x_4D(ix,iy,iz,iyloc,N_particle,sp1,sp1_x)
c------------------------------------------------------------------------
c    Master Process Prints Data
c------------------------------------------------------------------------
      if (Master) then
      T7 = MPI_WTIME()
        if (mod(it,iprnt).eq.0) then
           call prtout(idate,ut,ix,iy,iz,sg1,sl1,iout)
        endif
        if (mod(it,iscrat).eq.0) then 
           call prtemp(idate,ut,ix,iy,iz,sg1,sl1,iout)
        endif
      T8 = MPI_WTIME()
      endif
c     
10    continue  ! end of Big Loop
c-----------------------------------------------------------------------c
c                      Simulation ends here                             c
c-----------------------------------------------------------------------c

c      call close_ioap
c-------------------------------------------------------------------------------
c    Stop the timer
c-------------------------------------------------------------------------------
      EndTime = MPI_WTIME()-StartTime
      if (Master) then
         print*,'Master prints ascii files ...'
c         call plot3d(Nprocs,EndTime,ut,ix,iy,iz,iLm,ixtrn,
c     &                iytrn,iztrn,irxng,sg1)
         open(unit=98,file='/scratch/pmiehe/output/xypartimes.m',
     &              action='WRITE',position='APPEND');
         write(98,*) 'Nprocs=',Nprocs,'  CPU_TIME=',EndTime,' sec.'
         write(98,*) 'IO-time=',T4-StartTime+T8-T7,
     &                'distr-time=',T6-T5	 
         close(98)
      endif
      print("('Proc[',I2,'] Time = ',F9.2,' sec. = ',F7.2,' min.')"),
     &      MyId,EndTime,EndTime/60.0    
c        
      call MPI_FINALIZE( Ierr )
c      
      end program aq_driver_xy



c********************************************************************
      subroutine cho_tmp(ix,iy,iz,num,sg1,sx)
c********************************************************************
      parameter (mxspg=250)
      include 'aqindx.cmm'
      dimension sg1(ix,iy,iz,*),sx(iy,iz,2,*)
      do i=1,ix
      do j=1,iy
      do k=1,iz
      sg1(i,j,k,7)=40.*sg1(i,j,k,iair)*1.e-09
      enddo
      enddo
      enddo
      return
      end


      subroutine plot3d(Np,cpu,ut,ix,iy,iz,is,
     &                ixtrn,iytrn,iztrn,irxng,sg1)
      integer Np,ix,iy,iz,is,ixtrn,iytrn,iztrn,irxng
      real sg1(ix,iy,iz,is)
      double precision cpu
            integer ilayer
      character(len=4) :: c
      do ilayer=1,iz!,iz-1
      write(c,10) 1000+ilayer
      open(10,file=
     & '/scratch/pmiehe/output/parpgf_lyr' // c(3:4) // '.m')
      write(10,*) '%******** Parallel Symmetric Code **************'
      write(10,*) 'Nprocs=',Np,';'     
      write(10,*) 'P_CPU_TIME=',cpu,';'    
      write(10,*) 'P_ut=',ut,';'     
      write(10,*) 'P_ixtrn=',ixtrn,';'
      write(10,*) 'P_iytrn=',iytrn,';'
      write(10,*) 'P_iztrn=',iztrn,';'
      write(10,*) 'P_irxng=',irxng,';'
      write(10,*) 'P_PGF=['
      write(10,11) (((sg1(i,j,ilayer,l),i=1,ix),j=1,iy),l=1,27)
      write(10,*) '];'
      write(10,*) '%***********************************'
      close(10)
      enddo
 10   format(I4)
 11   format(9(E16.8,1X))     
      return
      end
